/*
Copyright 2015-2016, Manuel Bichler
WWW: <http://dbai.tuwien.ac.at/proj/lpopt/>

This file is part of lpopt.

lpopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

lpopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with lpopt.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <algorithm>
#include <memory>
#include <climits>
#include <map>

#include "asp-domain.hpp"
#include "safetyCharacterization.hpp"

using namespace std;

namespace domain {
  
unsigned int count_union(varset::iterator lbegin, varset::iterator lend, varset::iterator rbegin, varset::iterator rend);

template <typename T>
set<BodyElement *>* selectSafe(const Variable& v, const varset& alreadySafe, const varset& globVars, const SafetyCharacterization* bodySafety,  const vector<T *> *bodyElements);

template <typename T>
vector<const BodyElement *>* selectSafeVector(const Variable& v, const varset& alreadySafe, const varset& globVars, const SafetyCharacterization* bodySafety, const vector<T *> *bodyElements);

template <typename T>
vector<BodyElement *>* selectSafeVectorNoconst(const Variable& v, const varset& alreadySafe, const varset& globVars, const SafetyCharacterization* bodySafety, const vector<T *> *bodyElements);

template <typename T>
vector<const BodyElement *>* selectSafeVector(const Variable& v, const varset& globVars, SafetyCharacterization* bodySafety, const vector<T *> *bodyElements); // alreadySafe is empty, deletes bodySafety.

template <typename T>
vector<BodyElement *>* selectSafeVectorNoconst(const Variable& v, const varset& globVars, SafetyCharacterization* bodySafety, const vector<T *> *bodyElements); // alreadySafe is empty, deletes bodySafety.

void eraseFrom(varset& s1, const varset& v2); // erases all variables in v2 from v1

unsigned int count_union(varset::iterator first1, varset::iterator last1, varset::iterator first2, varset::iterator last2)
{
  unsigned int count = 0;
  while (true)
  {
    if (first1==last1 && first2==last2) return count;
    else if (first1==last1) {
      ++first2;
    }
    else if (first2==last2) {
      ++first1;
    }
    else if (*first1<*first2)
      ++first1;
    else if (*first2<*first1)
      ++first2;
    else {
      ++first1;
      ++first2;
    }
    ++count;
  }
}

template <typename T>
vector<BodyElement *>* selectSafeVectorNoconst(const Variable& v, const varset& globVars, SafetyCharacterization* bodySafety, const vector<T *> *bodyElements)
{
  varset* empty = new varset();
  vector<BodyElement *>* ret = selectSafeVectorNoconst(v, *empty, globVars, *bodySafety, bodyElements);
  delete empty;
  delete bodySafety;
  return ret;
}

template <typename T>
vector<const BodyElement *>* selectSafeVector(const Variable& v, const varset& globVars, SafetyCharacterization* bodySafety, const vector<T *> *bodyElements)
{
  varset* empty = new varset();
  vector<const BodyElement *>* ret = selectSafeVector(v, *empty, globVars, *bodySafety, bodyElements);
  delete empty;
  delete bodySafety;
  return ret;
}

template <typename T>
vector<BodyElement *>* selectSafeVectorNoconst(const Variable& v, const varset& alreadySafe, const varset& globVars, const SafetyCharacterization& bodySafety, const vector<T *> *bodyElements)
{
  set<BodyElement *>* set = selectSafe(v, alreadySafe, globVars, bodySafety, bodyElements);
  vector<BodyElement *>* ret = new vector<BodyElement *>(set->begin(), set->end());
  delete set;
  return ret;
}

template <typename T>
vector<const BodyElement *>* selectSafeVector(const Variable& v, const varset& alreadySafe, const varset& globVars, const SafetyCharacterization& bodySafety, const vector<T *> *bodyElements)
{
  set<BodyElement *>* set = selectSafe(v, alreadySafe, globVars, bodySafety, bodyElements);
  vector<const BodyElement *>* ret = new vector<const BodyElement *>(set->begin(), set->end());
  delete set;
  return ret;
}

void eraseFrom(varset& s1, const varset& v2)
{
  varset::iterator it1 = s1.begin();
  varset::iterator varit = v2.begin();
  while(it1 != s1.end() && varit != v2.end())
  {
    if(var_cmp()(*it1, *varit))
      ++it1;
    else if (var_cmp()(*varit, *it1))
      ++varit;
    else { // equal
      it1 = s1.erase(it1);
      ++varit;
    }
  }
}

template <typename T>
set<BodyElement *>* selectSafe(const Variable& v, const varset& alreadySafe, const varset& globVars, const SafetyCharacterization& bodySafety, const vector<T *> *bodyElements)
{
	static_assert(
			std::is_same<BodyElement, T>::value || std::is_same<NafLiteral, T>::value,
			"T must be a either BodyElement or NafLiteral"
	);
	
	if (alreadySafe.find(&v) != alreadySafe.end())
		return new set<BodyElement *>();

	// search for bodyElements that can make v safe
	map<unsigned int, vector<pair<BodyElement*, unique_ptr<SafetyCharacterization>>>> candidateElements; // key: number of not-already-safe variables of the body element
	for (T* it : *bodyElements) {
		const SafetyCharacterization* constst = it->safety(globVars);
		SafetyCharacterization* sc = constst->clone();
		it->dispose(constst);
		sc->addSafe(alreadySafe);
		if(sc->saveable(&v))
		{
			unique_ptr<varset> vars(((BodyElement*)it)->variables());
			eraseFrom(*vars, alreadySafe);
			candidateElements[vars->size()].push_back(pair<BodyElement*, unique_ptr<SafetyCharacterization>>(it, unique_ptr<SafetyCharacterization>(sc)));
		}
		else
		  delete sc;
	}
	// now that they are sorted, try them:
	set<BodyElement *>* bestSafeBody = NULL;
	unsigned int numberOfVariables = UINT_MAX;
	for (map<unsigned int, vector<pair<BodyElement*, unique_ptr<SafetyCharacterization>>>>::iterator mit = candidateElements.begin(); mit != candidateElements.end(); ++mit)
	{
	  if (mit->first >= numberOfVariables) break; // coming candidates have more elements than an already found safeBody
	  vector<pair<BodyElement*, unique_ptr<SafetyCharacterization>>>& candidates = mit->second;
	  for (vector<pair<BodyElement*, unique_ptr<SafetyCharacterization>>>::iterator cit = candidates.begin(); cit != candidates.end(); ++cit)
	  {
	    unique_ptr<varset> vars(cit->first->variables());
	    eraseFrom(*vars, alreadySafe);
	    
	    set<varset*>* neededs (cit->second->savingCandidates(&v)); // a set of all those (minimal) varsets that have the property: if safe, v would also be safe.
	    // for each, try to make them safe:
	    for (set<varset*>::iterator nit = neededs->begin(); nit != neededs->end(); ++nit)
	    {
	      if (count_union(vars->begin(), vars->end(), (*nit)->begin(), (*nit)->end()) > numberOfVariables)
		continue;
	      // try to make all the variables in this set safe:
	      set<BodyElement *>* candidateSafeBody = new set<BodyElement *>();
	      candidateSafeBody->insert(cit->first);
	      SafetyCharacterization cumulativeSafe (new varset(alreadySafe), true);
	      cumulativeSafe.incorporate(*(cit->second));
	      varset& toMakeSafe = **nit;
	      while (!toMakeSafe.empty())
	      {
		// begin with a variable that could be made safe without adding new rules:
		varset::iterator selectedVar = toMakeSafe.begin();
		while(selectedVar != toMakeSafe.end())
		  if (bodySafety.isSafe(*selectedVar, cumulativeSafe.safeVars()))
		    break;
		  else
		    ++selectedVar;
		if (selectedVar == toMakeSafe.end())
		// no variable that could be made safe without adding new rules
		// just take first one TODO maybe more efficient approach possible
		  selectedVar = toMakeSafe.begin();
		set<BodyElement *>* retRes = selectSafe(**selectedVar, cumulativeSafe.safeVars(), globVars, bodySafety, bodyElements);
		candidateSafeBody->insert(retRes->begin(), retRes->end());
		for(set<BodyElement *>::iterator it = retRes->begin(); it != retRes->end(); ++it)
		{
		  const SafetyCharacterization* sc = (*it)->safety(globVars);
		  cumulativeSafe.incorporate(*sc);
		  (*it)->dispose(sc);
		}
		delete retRes;
		toMakeSafe.erase(selectedVar);
	      }
	      delete *nit;
	      // now, we have a fully-featured candidateSafeBody. temporarily save it if its variable count is better than the current candidate
	      if(cumulativeSafe.safeVars().size() < numberOfVariables)
	      {
		numberOfVariables = cumulativeSafe.safeVars().size();
		delete bestSafeBody;
		bestSafeBody = candidateSafeBody;
	      }
	      else
		delete candidateSafeBody;
	    }
	    delete neededs;
	  }
	}
	return bestSafeBody;
}

SafetyCharacterization* TermTuple::safety() const {
	if (terms->size() == 1)
		return (*terms)[0]->safety();
	SafetyCharacterization* sc = new SafetyCharacterization();
	for (vector<Term *>::iterator it = terms->begin(); it != terms->end();
			++it) {
		sc->combine((*it)->safety());
	}
	return sc;
}

SafetyCharacterization* TermTuples::safety() const {
	if (tuples->size() != 1)
		return NULL; /*TODO: error/warning*/
	return (*tuples)[0]->safety();
}

SafetyCharacterization* UnaryTerm::safety() const {
	if (type == NEGATION)
		return base->safety();
	else
		return new SafetyCharacterization(base->variables(), false);
}

SafetyCharacterization* BracketedTerm::safety() const {
	return base->safety();
}

SafetyCharacterization* BinaryTerm::safety() const {
	return (new SafetyCharacterization(left->variables(), false))->combine(
			new SafetyCharacterization(right->variables(), false)); // all unsafe
}

SafetyCharacterization* PoolingTerm::safety() const { /*TODO: error/warning */
	return NULL;
}

SafetyCharacterization* Constant::safety() const {
	return new SafetyCharacterization();
}

SafetyCharacterization* Variable::safety() const {
	return new SafetyCharacterization(this, true);
}

SafetyCharacterization* AnonymousVariable::safety() const {
	return new SafetyCharacterization();
}

SafetyCharacterization* Function::safety() const {
	return tuples->safety();
}

NafLiteral::~NafLiteral() {
	delete safetyCh;
}

NafLiteral::NafLiteral(NafLiteral const &that) : safetyCh(that.safetyCh==NULL ? NULL : new SafetyCharacterization(*(that.safetyCh))) {
}

const SafetyCharacterization* NafLiteral::safety(const varset& /*globalVars*/) {
	if (safetyCh == NULL)
		safetyCh = computeSafetyCharacterization();
	return safetyCh;
}

void NafLiteral::dispose(const SafetyCharacterization* /*safety*/) const {
}

const SafetyCharacterization* ClassicalLiteral::computeSafetyCharacterization() const
{
	if (nafNegated)
		return new SafetyCharacterization(((BodyElement*)(this))->variables(), false); // all variables unsafe
	if (tuples == NULL)
		return new SafetyCharacterization();
	return tuples->safety();
}

SafetyCharacterization* LiteralBody::safety(const varset& globalVars) const
{
	SafetyCharacterization* sc = new SafetyCharacterization();
	for (vector<NafLiteral *>::iterator it = literals->begin(); it != literals->end();
			++it) {
		const SafetyCharacterization* tmp = (*it)->safety(globalVars);
		sc->incorporate(*tmp);
		(*it)->dispose(tmp);
	}
	return sc;
}
SafetyCharacterization* LiteralBody::safety() const
{
	SafetyCharacterization* sc = new SafetyCharacterization();
	for (vector<NafLiteral *>::iterator it = literals->begin(); it != literals->end();
			++it) {
		const SafetyCharacterization* tmp = (*it)->safety(globalVariables);
		sc->incorporate(*tmp);
		(*it)->dispose(tmp);
	}
	return sc;
}
vector<const BodyElement *>* LiteralBody::unsafe() const
{
	vector<const BodyElement *>* ret = new vector<const BodyElement *>();
	for (vector<NafLiteral *>::iterator it = literals->begin(); it != literals->end(); ++it) {
		const SafetyCharacterization* tmp = (*it)->safety(globalVariables);
		if (!tmp->unsafeVars().empty())
			ret->push_back(*it);
		delete tmp;
	}
	return ret;
}
vector<const BodyElement *>* LiteralBody::selectSafe(const Variable& v) const
{
	return selectSafeVector(v, globalVariables, safety(), literals);
}

vector<BodyElement *>* LiteralBody::selectSafe(const Variable& v)
{
	return selectSafeVectorNoconst(v, globalVariables, safety(), literals);
}

SafetyCharacterization* OptimizeElement::safety() const
{
	unique_ptr<varset> globVars(variables());
	return literals->safety(*globVars);
}
vector<const BodyElement *>* OptimizeElement::unsafe() const
{
	return literals->unsafe();
}
vector<const BodyElement *>* OptimizeElement::selectSafe(const Variable& v) const
{
	const auto * l = literals;
	return l->selectSafe(v);
}

vector<BodyElement *>* OptimizeElement::selectSafe(const Variable& v)
{
	return literals->selectSafe(v);
}

SafetyCharacterization* OptimizeStatement::safety() const
{
	if (elements->empty()) return new SafetyCharacterization();
	if (elements->size() > 1) return NULL; //TODO error/warning
	return (*elements)[0]->safety();
}
vector<const BodyElement *>* OptimizeStatement::unsafe() const
{
	if (elements->size() == 0)
		return new vector<const BodyElement *>();
	vector<const BodyElement *>* ret = (*elements)[0]->unsafe();
	for (vector<OptimizeElement *>::iterator it = ++elements->begin(); it != elements->end(); ++it)
	{
		vector<const BodyElement *>* part = (*it)->unsafe();
		ret->insert(ret->end(), part->begin(), part->end());
		delete part;
	}
	return ret;
}
vector<const BodyElement *>* OptimizeStatement::selectSafe(const Variable& v) const
{
	if (elements->size() != 1)
		return NULL; //TODO error/warning
	const auto * e = (*elements)[0];
	return e->selectSafe(v);
}
vector<BodyElement *>* OptimizeStatement::selectSafe(const Variable& v)
{
	if (elements->size() != 1)
		return NULL; //TODO error/warning
	return (*elements)[0]->selectSafe(v);
}

SafetyCharacterization* Body::safety() const
{
	return safety(globalVariables);
}
SafetyCharacterization* Body::safety(const varset &globalVars) const
{
	SafetyCharacterization* sc = new SafetyCharacterization();
	for (vector<BodyElement *>::iterator it = elements->begin(); it != elements->end();
			++it) {
		const SafetyCharacterization* tmp = (*it)->safety(globalVars);
		sc->incorporate(*tmp);
		(*it)->dispose(tmp);
	}
	return sc;
}
vector<const BodyElement *>* Body::unsafe() const
{
	vector<const BodyElement *>* ret = new vector<const BodyElement *>();
	for (vector<BodyElement *>::iterator it = elements->begin(); it != elements->end(); ++it) {
		const SafetyCharacterization* tmp = (*it)->safety(globalVariables);
		if (!tmp->unsafeVars().empty())
			ret->push_back(*it);
		delete tmp;
	}
	return ret;
}
vector<const BodyElement *>* Body::selectSafe(const Variable& v) const
{
	return selectSafeVector(v, globalVariables, safety(), elements);
}

vector<BodyElement *>* Body::selectSafe(const Variable& v)
{
	return selectSafeVectorNoconst(v, globalVariables, safety(), elements);
}

SafetyCharacterization* WeakConstraint::safety() const
{
	return body->safety();
}
vector<const BodyElement *>* WeakConstraint::unsafe() const
{
	return body->unsafe();
}
vector<const BodyElement *>* WeakConstraint::selectSafe(const Variable& v) const
{
	const auto * b = body;
	return b->selectSafe(v);
}

vector<BodyElement *>* WeakConstraint::selectSafe(const Variable& v)
{
	return body->selectSafe(v);
}

SafetyCharacterization* Rule::safety() const
{
	return body->safety();
}
vector<const BodyElement *>* Rule::unsafe() const
{
  return body->unsafe();
}
vector<const BodyElement *>* Rule::selectSafe(const Variable& v) const
{
  const auto * b = body;
  return b->selectSafe(v);
}

vector<BodyElement *>* Rule::selectSafe(const Variable& v)
{
  return body->selectSafe(v);
}

SafetyCharacterization* Guard::safety() const
{
  return term->safety();
}

const SafetyCharacterization* Aggregate::safety(const varset& globalVars)
{
  varset* insideVars = insideVariables();
  // only those globally bound
  vector<const Variable*> insideBound(insideVars->size());
  vector<const Variable*>::iterator it = std::set_intersection(insideVars->begin(), insideVars->end(), globalVars.begin(), globalVars.end(), insideBound.begin(), var_cmp());
  insideBound.resize(it - insideBound.begin());
  insideVars->clear();
  insideVars->insert(insideBound.begin(), insideBound.end());
  
  SafetyCharacterization *sc = new SafetyCharacterization(insideVars, false);
  
  if (nafNegated) // all inner bound and guard variables are unsafe.
  {
    varset* guardVars = ((BodyElement*)this)->variables();
    sc->addUnsafe(*guardVars);
    delete guardVars;
    return sc;
  }
  
  // no "not". consider guards:
  SafetyCharacterization *lsc=NULL, *rsc=NULL;
  if(guard_l != NULL)
  {
    if (guard_l->equal())
    {
      lsc = guard_l->safety();
      lsc->equalsIncorporate(*sc);
    }
    else
    { // guards variables of type !=, <, >, <=, >= are unsafe
      lsc = new SafetyCharacterization(guard_l->variables(), false);
      lsc->incorporate(*sc);
    }
  }
  if(guard_r != NULL)
  {
    if (guard_r->equal())
    {
      rsc = guard_r->safety();
      rsc->equalsIncorporate(*sc);
    }
    else
    { // guards variables of type !=, <, >, <=, >= are unsafe
      rsc = new SafetyCharacterization(guard_r->variables(), false);
      rsc->incorporate(*sc);
    }
  }
  
  if (lsc == NULL && rsc == NULL)
    return sc;
  delete sc;
  if (lsc == NULL)
    return rsc;
  if (rsc == NULL)
    return lsc;
  
  return lsc->combine(rsc);
}

void Aggregate::dispose(const SafetyCharacterization* safety) const
{
  delete safety;
}

const SafetyCharacterization* BuiltinAtom::computeSafetyCharacterization() const
{
  if (!equal()) return new SafetyCharacterization(((BodyElement*)this)->variables(), false); // all unsafe
  return t1->safety()->equalsCombine(t2->safety());
}

LiteralBody* LiteralBody::splitContains(const varset& vars)
{
	vector<NafLiteral *>* newBody = new vector<NafLiteral *>();
	vector<NafLiteral *>::iterator it = literals->begin();
	while(it != literals->end())
	{
		varset* litvars = ((BodyElement*)(*it))->variables();
		if (empty_intersection(*litvars, vars))
		{
		  newBody->push_back(*it);
		  it = literals->erase(it);
		}
		else
		  ++it;
		delete litvars;
	}
	
	LiteralBody* newLitBody = new LiteralBody(newBody);
	const SafetyCharacterization* s;
	while(!(s = newLitBody->safety())->unsafeVars().empty())
	{
		it = newBody->begin();
		while (it != newBody->end())
		{
			if (!empty_intersection(s->unsafeVars(), *((*it)->variables(vars /*actually, parameter is being ignored*/))))
			{ // get it back into this!
				this->literals->push_back(*it);
				it = newBody->erase(it);
			}
			else
			{
				++it;
			}
		}
	}
	
	calcVariables();
	newLitBody->calcVariables();
	return newLitBody;
}
bool empty_intersection(const varset& x, const varset& y) // returns true if no intersection, false otherwise.
{
    varset::const_iterator i = x.begin();
    varset::const_iterator j = y.begin();
    while (i != x.end() && j != y.end())
    {
      if (**i == **j)
        return false;
      else if (**i < **j)
        ++i;
      else
        ++j;
    }
    return true;
}



} // domain

 
