/*
Copyright 2015-2016, Manuel Bichler
WWW: <http://dbai.tuwien.ac.at/proj/lpopt/>

This file is part of lpopt.

lpopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

lpopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with lpopt.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SAFETY_CHARACTERIZATION_H_
#define SAFETY_CHARACTERIZATION_H_

#include<set>
#include<string>

#include "asp-domain.hpp"

using namespace std;

namespace domain
{

  class SafetyRule;
  class SafetyCharacterization;

  struct rule_cmp // functor comparing two SafetyRule pointers
  {
    bool operator() (SafetyRule * lhs, SafetyRule * rhs) const;    
  };

  typedef set<SafetyRule*, rule_cmp> ruleset;

  
  class SafetyRule // represents an induction rule of form A1,A2,...,An -> B. semantics: if all Ai are safe, B is safe.
  {
  public:
    SafetyRule(varset* dependees, const Variable* depender);
  public:
    ~SafetyRule();
    SafetyRule(SafetyRule const &that);
    SafetyRule& operator=(SafetyRule that);
    SafetyRule(SafetyRule&& other);
    friend void swap(SafetyRule& first, SafetyRule& second);
    friend bool operator==(const SafetyRule& lhs, const SafetyRule& rhs);
    friend bool operator<(const SafetyRule& lhs, const SafetyRule& rhs);
    SafetyRule* clone() const;
  public:
    const Variable* perform(const varset& safes) const; // returns the depender if the safes trigger this rule, otherwise returns NULL.
    const Variable& getDepender() const; //returns the depender
    const varset& getDependees() const; //returns the dependees
    bool includes(const SafetyRule &other) const; // returns true if this rule includes the other rule (i.e. if the dependers are equal and this dependees are a subset of the other dependees)
    void incorporate(const SafetyRule &other); // if the other rule's dependees are a subset of this rule's dependees, removes the other rule's depender from this rule's dependees.
    void removeDependees(const varset& vars); // removes vars from the dependees set
  private:
    SafetyRule() {}
  private:
    varset* dependees;
    const Variable* depender; // if the dependees are safe, the depender is also safe.
  };

  
  class SafetyCharacterization
  {
  public:
    SafetyCharacterization(); // empty constructor
    SafetyCharacterization(const Variable* var, bool safe); // constructor for single safe/unsafe variable
    SafetyCharacterization(varset* vars, bool safe); // all safe/unsafe. Uses the given varset. This object is responsible for its proper deletion. Do not delete yourself outside of this class.
  public:
    ~SafetyCharacterization();
    SafetyCharacterization(SafetyCharacterization const &that);
    SafetyCharacterization& operator=(SafetyCharacterization that);
    SafetyCharacterization(SafetyCharacterization&& other);
    friend void swap(SafetyCharacterization& first, SafetyCharacterization& second);
    SafetyCharacterization* clone() const;
  public: // the object always stays consistent: after each call, the intersection between safe and unsafe variables is empty, there are no rules that entail already safe variables, and there are no satisfied rules.
    void addRule(SafetyRule *rule);
    bool addSafe(const Variable* var); // adds if not already safe. returns whether it has been added.
    bool addUnsafe(const Variable* var); // adds if not safe or already unsafe. returns whether it has been added.
    void addSafe(const varset& vars);
    void addUnsafe(const varset& vars);
    bool add(const Variable* var, bool safe); // adds if not already inside. returns whether it has been added.
    void add(const varset& vars, bool safe);
    bool isSafe(const Variable* v) const;
    bool isSafe(const Variable* v, const varset& safes) const; // assuming safes are safe, returns true if v would then also be safe.
    const varset& safeVars() const;
    const varset& unsafeVars() const;
    varset* nonSafeable() const; // returns those unsafe variable that cannot be made safe by rules
    set<varset*>* savingCandidates(const Variable* v) const; // returns a set of all those (minimal) varsets that have the property: if safe, according to this characterization, v would also be safe.
    bool saveable(const Variable* v) const; // returns true if v is safe or if it is the depender of a rule
    ruleset* candidateRules(const Variable& unsafe) const; // returns a subset of this characterization's rules that might make the given variable safe
    SafetyCharacterization* combine(SafetyCharacterization* other); // incorporates other, deletes other, returns this
    SafetyCharacterization* incorporate(const SafetyCharacterization& other); // incorporates other, returns this
    SafetyCharacterization* equalsCombine(SafetyCharacterization* other); // combines into this the safety characterization of an atom "t1 = t2" where this is one term's safety characterization and other is the other's. deletes other, returns this. Neither this nor other must have rules. Afterwards, this might have rules.
    SafetyCharacterization* equalsIncorporate(const SafetyCharacterization& other); // incorporates into this the safety characterization of an atom "t1 = t2" where this is one term's safety characterization and other is the other's. returns this. Neither this nor other must have rules. Afterwards, this might have rules.
  private:
    void rectifyAfterAddSafes(); // triggers the rules until no more possible. then deletes the safe from unsafe. Deletes rules that are already fulfilled (i.e. whose dependers are already safe).
    void forceAddUnsafe(const varset& vars); // just throws the given vars into the unsafe set. no integrity checking. perform rectifyAfterAddSafes after this call.
  private:
    varset* safes;
    varset* unsafes;
    ruleset rules;
  };


} // namespace domain

#endif // SAFETY_CHARACTERIZATION_H_
