/*
Copyright 2015-2016, Manuel Bichler
WWW: <http://dbai.tuwien.ac.at/proj/lpopt/>

This file is part of lpopt.

lpopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

lpopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with lpopt.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <algorithm>
#include <memory>

#include "asp-domain.hpp"

using namespace std;
using namespace domain;

#define SETINIT varset *result = new varset()

varset* TermTuple::variables() const
{
  SETINIT;
  for(vector<Term *>::iterator it = terms->begin(); it!=terms->end(); ++it)
  {
    varset *partSet = (*it)->variables();
    result->insert(partSet->begin(), partSet->end());
    delete partSet;
  }
  return result;
}

varset* TermTuples::variables() const
{
  SETINIT;
  for(vector<TermTuple *>::iterator it = tuples->begin(); it!=tuples->end(); ++it)
  {
    varset *partSet = (*it)->variables();
    result->insert(partSet->begin(), partSet->end());
    delete partSet;
  }
  return result;
}

varset* UnaryTerm::variables() const
{
  return base->variables();
}

varset* BracketedTerm::variables() const
{
  return base->variables();
}

varset* BinaryTerm::variables() const
{
  varset* leftv = left->variables();
  varset* rightv = right->variables();
  leftv->insert(rightv->begin(), rightv->end());
  delete rightv;
  return leftv;
}

varset* PoolingTerm::variables() const
{
  SETINIT;
  for(vector<Term *>::iterator it = terms->begin(); it!=terms->end(); ++it)
  {
    varset *partSet = (*it)->variables();
    result->insert(partSet->begin(), partSet->end());
    delete partSet;
  }
  return result;
}

varset* Constant::variables() const
{
  return new varset();
}

varset* Variable::variables() const
{
  SETINIT;
  result->insert(this);
  return result;
}

varset* AnonymousVariable::variables() const
{
  return new varset();
}

varset* Function::variables() const
{
  return tuples->variables();
}

void ClassicalLiteral::updatePublicVariables()
{
  publicVariables.clear();
  if (tuples)
  {
    varset* tvs = tuples->variables();
    publicVariables.insert(tvs->begin(), tvs->end());
    delete tvs;
  }
}

varset* WeightAtLevel::variables() const
{
  varset* result = weight->variables();
  if (level) {
    varset *partSet = level->variables();
    result->insert(partSet->begin(), partSet->end());
    delete partSet;
  }
  if (tuple) {
    varset *partSet = tuple->variables();
    result->insert(partSet->begin(), partSet->end());
    delete partSet;
  }
  return result;
}

varset* OptimizeElement::variables() const
{
  return literals->variables();
}
varset* OptimizeElement::headVars() const
{
  return weightAtLevel->variables();
}
set<varset*>* OptimizeElement::elementVars() const
{
  set<varset*>* ret = literals->elementVars();
  varset* globs = variables();
  varset* head = headVars();

  varset::iterator headit = head->begin();
  varset::iterator globsit = globs->begin();
  while (headit != head->end() && globsit != globs->end())
  {
    if (var_cmp()(*headit, *globsit)) // *headit < *globsit
      headit = head->erase(headit);
    else
    {
      if (var_cmp()(*globsit, *headit)) // *headit > *globsit
	++globsit;
      else // *headit == *globsit
      {
	++headit;
	++globsit;
      }
    }
  }
  head->erase(headit, head->end()); // if globsit is on its end, delete rest elements of head
  
  delete globs;
  ret->insert(head);
  return ret;
}
set<varset*>* OptimizeElement::bodyVars() const
{
  return literals->elementVars();
}
vector<const BodyElement *>* OptimizeElement::selectBody(varset &s) const
{
  const auto* l = literals;
  return l->select(s);
}
vector<BodyElement *>* OptimizeElement::selectBody(varset &s)
{
  return literals->select(s);
}
vector<BodyElement *>* OptimizeElement::extractBody(varset &s)
{
  return literals->extract(s);
}
vector<const BodyElement *>* OptimizeElement::variableFreeBody() const
{
  return literals->variableFree();
}

varset* OptimizeStatement::variables() const
{
  SETINIT;
  for(vector<OptimizeElement *>::iterator it = elements->begin(); it!=elements->end(); ++it)
  {
    varset *partSet = (*it)->variables();
    result->insert(partSet->begin(), partSet->end());
    delete partSet;
  }
  return result;
}
varset* OptimizeStatement::headVars() const
{
  auto size = elements->size();
  if (size == 0)
    return new varset();
  if (size > 1)
    return NULL; // TODO warning/error
  return (*elements)[0]->headVars();
}
set<varset*>* OptimizeStatement::elementVars() const
{
  auto size = elements->size();
  if (size == 0)
    return new set<varset*>();
  if (size > 1)
    return NULL; // TODO warning/error
  return (*elements)[0]->elementVars();
}
set<varset*>* OptimizeStatement::bodyVars() const
{
  auto size = elements->size();
  if (size == 0)
    return new set<varset*>();
  if (size > 1)
    return NULL; // TODO warning/error
  return (*elements)[0]->bodyVars();

}
vector<const BodyElement *>* OptimizeStatement::selectBody(varset &s) const
{
  auto size = elements->size();
  if (size == 0)
    return new vector<const BodyElement *>();
  if (size > 1)
    return NULL; // TODO warning/error
  const auto *e = (*elements)[0];
  return e->selectBody(s);
}
vector<BodyElement *>* OptimizeStatement::selectBody(varset &s)
{
  auto size = elements->size();
  if (size == 0)
    return new vector<BodyElement *>();
  if (size > 1)
    return NULL; // TODO warning/error
  return (*elements)[0]->selectBody(s);
}
vector<BodyElement *>* OptimizeStatement::extractBody(varset &s)
{
  auto size = elements->size();
  if (size == 0)
    return new vector<BodyElement *>();
  if (size > 1)
    return NULL; // TODO warning/error
  return (*elements)[0]->extractBody(s);
}
vector<BodyElement *>* OptimizeStatement::extractBodyWithoutGlobvarUpdate(varset &s)
{
  return extractBody(s);
}
void OptimizeStatement::calcVariables()
{
}
vector<const BodyElement *>* OptimizeStatement::variableFreeBody() const
{
  if (elements->size() == 0)
    return new vector<const BodyElement *>();
  vector<const BodyElement *>* ret = (*elements)[0]->variableFreeBody();
  for (vector<OptimizeElement *>::iterator it = ++elements->begin(); it != elements->end(); ++it)
  {
    vector<const BodyElement *>* part = (*it)->variableFreeBody();
    ret->insert(ret->end(), part->begin(), part->end());
    delete part;
  }
  return ret;
}

void Body::calcVariables()
{
  if (elements->empty())
    globalVariables.clear();
  else
  {
    vector<BodyElement *>::iterator it = elements->begin();
    varset* evs = (*it)->variables();
    globalVariables.swap(*evs);
    delete evs;
    for(++it; it!=elements->end(); ++it)
    {
      unique_ptr<varset> partSet ((*it)->variables());
      globalVariables.insert(partSet->begin(), partSet->end());
    }
  }
}
varset* Body::variables() const
{
  return new varset(globalVariables);
}
set<varset*>* Body::elementVars() const
{
  set<varset*> *ret = new set<varset*>();
  for(vector<BodyElement *>::iterator it = elements->begin(); it!=elements->end(); ++it)
    ret->insert((*it)->variables(globalVariables));
  return ret;
}
vector<const BodyElement *>* Body::select(varset &s) const
{
  vector<const BodyElement *>* ret = new vector<const BodyElement *>();
  for(vector<BodyElement *>::iterator it = elements->begin(); it!=elements->end(); ++it)
  {
    unique_ptr<varset> elglobs((*it)->variables(globalVariables));
    if (elglobs->empty())
      continue;
    if (std::includes(s.begin(), s.end(), elglobs->begin(), elglobs->end(), var_cmp()))
      ret->push_back(*it);
  }
  return ret;
}
vector<BodyElement *>* Body::select(varset &s)
{
  vector<BodyElement *>* ret = new vector<BodyElement *>();
  for(vector<BodyElement *>::iterator it = elements->begin(); it!=elements->end(); ++it)
  {
    unique_ptr<varset> elglobs((*it)->variables(globalVariables));
    if (elglobs->empty())
      continue;
    if (std::includes(s.begin(), s.end(), elglobs->begin(), elglobs->end(), var_cmp()))
      ret->push_back(*it);
  }
  return ret;
}
vector<BodyElement *>* Body::extract(varset &s)
{
  auto ret = extractWithoutGlobvarUpdate(s);
  calcVariables();
  return ret;
}

vector<BodyElement *>* Body::extractWithoutGlobvarUpdate(varset &s)
{
  auto ret = new vector<BodyElement *>();
  auto it = elements->begin();
  while(it!=elements->end())
  {
    if ((*it)->globalVarsNotEmptyAndIncludedIn(globalVariables, s))
    {
      ret->push_back(*it);
      it = elements->erase(it);
    }
    else
      ++it;
  }
  return ret;
}
vector<const BodyElement *>* Body::variableFree() const
{
  vector<const BodyElement *>* ret = new vector<const BodyElement *>();
  for(vector<BodyElement *>::iterator it = elements->begin(); it!=elements->end(); ++it)
  {
    unique_ptr<varset> elglobs((*it)->variables(globalVariables));
    if (elglobs->empty())
      ret->push_back(*it);
  }
  return ret;
}

varset* WeakConstraint::variables() const
{
  return body->variables();
}
varset* WeakConstraint::headVars() const
{
  return weightAtLevel->variables(); // w@l must not contain non-global variables, otherwise unsafe.
}
set<varset*>* WeakConstraint::elementVars() const
{
  set<varset*>* ret = body->elementVars();
  ret->insert(headVars());
  return ret;
}
set<varset*>* WeakConstraint::bodyVars() const
{
  set<varset*>* ret = body->elementVars();
  return ret;
}
vector<const BodyElement *>* WeakConstraint::selectBody(varset &s) const
{
  const auto *b = body;
  return b->select(s);
}
vector<BodyElement *>* WeakConstraint::selectBody(varset &s)
{
  return body->select(s);
}
vector<BodyElement *>* WeakConstraint::extractBody(varset &s)
{
  return body->extract(s);
}
vector<BodyElement *>* WeakConstraint::extractBodyWithoutGlobvarUpdate(varset &s)
{
  return extractBody(s);
}
void WeakConstraint::calcVariables()
{
}
vector<const BodyElement *>* WeakConstraint::variableFreeBody() const
{
  return body->variableFree();
}

varset* Rule::variables() const
{
  return body->variables();
}
varset* Rule::headVars() const {
  varset* headVars = head->variables();
  varset* global = variables();
  varset* intersection = new varset();
  std::set_intersection(headVars->begin(), headVars->end(), global->begin(), global->end(), inserter(*intersection, intersection->begin()), var_cmp());
  delete headVars;
  delete global;
  return intersection;
}
set<varset*>* Rule::elementVars() const
{
  varset* globals = body->variables();
  set<varset*>* ret = body->elementVars();
  ret->insert(head->variables(*globals));
  delete globals;
  return ret;
}
set<varset*>* Rule::bodyVars() const
{
  return body->elementVars();
}
vector<const BodyElement *>* Rule::selectBody(varset &s) const
{
  const auto *b = body;
  return b->select(s);
}
vector<BodyElement *>* Rule::selectBody(varset &s)
{
  return body->select(s);
}
vector<BodyElement *>* Rule::extractBody(varset &s)
{
  return body->extract(s);
}
vector<BodyElement *>* Rule::extractBodyWithoutGlobvarUpdate(varset &s)
{
  return body->extractWithoutGlobvarUpdate(s);
}
void Rule::calcVariables()
{
  body->calcVariables();
}
vector<const BodyElement *>* Rule::variableFreeBody() const
{
  return body->variableFree();
}

varset* Disjunction::variables() const
{
  SETINIT;
  for(vector<ClassicalLiteral *>::iterator it = literals->begin(); it!=literals->end(); ++it)
  {
    const varset& vars = ((BodyElement*)(*it))->variablesFast();
    result->insert(vars.begin(), vars.end());
  }
  return result;
}

varset* Guard::variables() const
{
  return term->variables();
}

varset* ChoiceElement::variables() const
{
  varset* result = dependencyLiterals->variables();
  const varset& partSet = ((BodyElement*)literal)->variablesFast();
  result->insert(partSet.begin(), partSet.end());
  return result;
}

varset* Choice::variables() const
{
  SETINIT;
  if(guard_l) {
    varset *partSet = guard_l->variables();
    result->insert(partSet->begin(), partSet->end());
    delete partSet;
  }
  if(guard_r) {
    varset *partSet = guard_r->variables();
    result->insert(partSet->begin(), partSet->end());
    delete partSet;
  }
  for(vector<ChoiceElement *>::iterator it = elements->begin(); it!=elements->end(); ++it)
  {
    varset *partSet = (*it)->variables();
    result->insert(partSet->begin(), partSet->end());
    delete partSet;
  }
  return result;
}

varset* AggregateElement::variables() const
{
  varset* result = dependencyLiterals->variables();
  if(termTuple)
  {
    varset* partSet = termTuple->variables();
    result->insert(partSet->begin(), partSet->end());
    delete partSet;
  }
  return result;
}

void Aggregate::updatePublicVariables()
{
  publicVariables.clear();
  if(guard_l) {
    varset *partSet = guard_l->variables();
    publicVariables.insert(partSet->begin(), partSet->end());
    delete partSet;
  }
  if(guard_r) {
    varset *partSet = guard_r->variables();
    publicVariables.insert(partSet->begin(), partSet->end());
    delete partSet;
  }
}

varset* Aggregate::insideVariables() const
{
  SETINIT;
  for(vector<AggregateElement *>::iterator it = elements->begin(); it!=elements->end(); ++it)
  {
    varset *partSet = (*it)->variables();
    result->insert(partSet->begin(), partSet->end());
    delete partSet;
  }
  return result;
}
varset* Aggregate::variables(const varset& globalVars) const
{
  varset* all = insideVariables();
  if(guard_l) {
    varset *partSet = guard_l->variables();
    all->insert(partSet->begin(), partSet->end());
    delete partSet;
  }
  if(guard_r) {
    varset *partSet = guard_r->variables();
    all->insert(partSet->begin(), partSet->end());
    delete partSet;
  }
  varset* ret = new varset();
  std::set_intersection(all->begin(), all->end(), globalVars.begin(), globalVars.end(), inserter(*ret, ret->begin()), var_cmp());
  delete all;
  return ret;
}
bool Aggregate::globalVarsNotEmptyAndIncludedIn(const varset& globalVariables, const varset& s) const
{
  unique_ptr<varset> elglobs(variables(globalVariables));
  return (!elglobs->empty() && std::includes(s.begin(), s.end(), elglobs->begin(), elglobs->end(), var_cmp()));
}

void BuiltinAtom::updatePublicVariables()
{
  publicVariables.clear();
  varset* leftv = t1->variables();
  publicVariables.insert(leftv->begin(), leftv->end());
  delete leftv;
  varset* rightv = t2->variables();
  publicVariables.insert(rightv->begin(), rightv->end());
  delete rightv;
}

void LiteralBody::calcVariables()
{
  if (literals->empty())
    globalVariables.clear();
  else
  {
    vector<NafLiteral *>::iterator it = literals->begin();
    varset* evs = ((BodyElement*)(*it))->variables();
    globalVariables.swap(*evs);
    delete evs;
    for(++it; it!=literals->end(); ++it)
    {
      const varset& partSet = ((BodyElement*)(*it))->variablesFast();
      globalVariables.insert(partSet.begin(), partSet.end());
    }
  }
}
varset* LiteralBody::variables() const
{
  return new varset(globalVariables);
}
set<varset*>* LiteralBody::elementVars() const
{
  set<varset*>* ret = new set<varset*>();
  for(vector<NafLiteral *>::iterator it = literals->begin(); it!=literals->end(); ++it)
    ret->insert(((BodyElement*)(*it))->variables());
  return ret;
}
vector<const BodyElement *>* LiteralBody::select(varset &s) const
{
  vector<const BodyElement *>* ret = new vector<const BodyElement *>();
  for(vector<NafLiteral *>::iterator it = literals->begin(); it!=literals->end(); ++it)
  {
    const varset& elglobs = ((BodyElement*)(*it))->variablesFast();
    if (elglobs.empty())
      continue;
    if (std::includes(s.begin(), s.end(), elglobs.begin(), elglobs.end(), var_cmp()))
      ret->push_back(*it);
  }
  return ret;
}
vector<BodyElement *>* LiteralBody::select(varset &s)
{
  vector<BodyElement *>* ret = new vector<BodyElement *>();
  for(vector<NafLiteral *>::iterator it = literals->begin(); it!=literals->end(); ++it)
  {
    const varset& elglobs = ((BodyElement*)(*it))->variablesFast();
    if (elglobs.empty())
      continue;
    if (std::includes(s.begin(), s.end(), elglobs.begin(), elglobs.end(), var_cmp()))
      ret->push_back(*it);
  }
  return ret;
}
vector<BodyElement *>* LiteralBody::extract(varset &s)
{
  vector<BodyElement *>* ret = extractWithoutGlobvarUpdate(s);
  calcVariables();
  return ret;
}
vector<BodyElement *>* LiteralBody::extractWithoutGlobvarUpdate(varset &s)
{
  vector<BodyElement *>* ret = new vector<BodyElement *>();
  vector<NafLiteral *>::iterator it = literals->begin();
  while(it!=literals->end())
  {
    const varset& elglobs = ((BodyElement*)(*it))->variablesFast();
    if (!elglobs.empty() && std::includes(s.begin(), s.end(), elglobs.begin(), elglobs.end(), var_cmp()))
    {
      ret->push_back(*it);
      it = literals->erase(it);
    }
    else
      ++it;
  }
  return ret;
}
vector<const BodyElement *>* LiteralBody::variableFree() const
{
  vector<const BodyElement *>* ret = new vector<const BodyElement *>();
  for(vector<NafLiteral *>::iterator it = literals->begin(); it!=literals->end(); ++it)
  {
    if ((*it)->variablesFast().empty())
      ret->push_back(*it);
  }
  return ret;
}

varset* NafLiteral::variables(const varset& /*globalVars*/) const
{
  return ((BodyElement*)this)->variables();
}
bool NafLiteral::globalVarsNotEmptyAndIncludedIn(const varset& /*globalVariables*/, const varset& s) const
{
  const varset& elglobs = ((BodyElement*)(this))->variablesFast();
  return (!elglobs.empty() && std::includes(s.begin(), s.end(), elglobs.begin(), elglobs.end(), var_cmp()));
}

varset* RuleHead::variables(const varset& globalVars) const
{
  varset* all = variables();
  varset* ret = new varset();
  std::set_intersection(all->begin(), all->end(), globalVars.begin(), globalVars.end(), inserter(*ret, ret->begin()), var_cmp());
  delete all;
  return ret;
}
