/*
Copyright 2015-2016, Manuel Bichler
WWW: <http://dbai.tuwien.ac.at/proj/lpopt/>

This file is part of lpopt.

lpopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

lpopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with lpopt.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <fstream>
#include <cstring>
#include <cstdlib>
#include <ctime>
#include <getopt.h>

#include <htd/main.hpp>
using namespace htd;

#include "optimization/LogicProgramOptimizationManager.hpp"

#include <parser/Driver.h>

using namespace std;
using namespace lpopt;

static void usage();

static char* sProgramName;

int main(int argc, char **argv)
{
//	clock_t startC = clock();

	sProgramName = argv[0];

	/* Argument Handling */

	int arg, argCount = 1;
	char *sEnd = NULL;
	bool sOpt = false, fOpt = false, hOpt = false, dOpt = false, iOpt = false, lOpt = false, vOpt = false;
	unsigned int seed = (unsigned int)time(NULL);
	DecompositionOrderingAlgorithm decomp = MIW;
	char *filename = NULL;
	char *logfileName = NULL;
	ifstream file;
	istream *stream = &std::cin;

	while((arg = getopt(argc, argv, "dvs:f:h:il:")) != EOF)
	{
		++argCount;
		switch(arg)
		{
		case 'h':
                        if(hOpt || !optarg) usage();
                        if(strcmp(optarg, "mcs") == 0) decomp = MCS;
                        else if(strcmp(optarg, "mf") == 0) decomp = MF;
                        else if(strcmp(optarg, "miw") == 0) decomp = MIW;
                        else if(strcmp(optarg, "noa") == 0) decomp = NOA;
                        else usage();
                        ++argCount;
                        hOpt = true;
                        break;
		case 's':
			if(sOpt || !optarg) usage();
			seed = (unsigned int)strtol(optarg, &sEnd, 0);
			if(*sEnd) usage();
			sOpt = true;
			++argCount;
			break;
		case 'd':
			if(dOpt) usage();
			dOpt = true;
			break;
		case 'v':
			if(vOpt) usage();
			vOpt = true;
			break;
		case 'f':
			if(fOpt || !optarg) usage();
			filename = optarg;
			fOpt = true;
			++argCount;
			break;
		case 'l':
			if(lOpt || !optarg) usage();
			logfileName = optarg;
			lOpt = true;
			++argCount;
			break;
		case 'i':
			if(iOpt) usage();
			iOpt = true;
			break;
		case '?':
			usage();
			break;
		default:
                	cerr << sProgramName << ": 'getopt' emitted an undocumented character!" << endl;
			exit(EXIT_FAILURE);
			break;
		}
	}

	if(argCount != argc) usage();

	if(vOpt)
	{
		cout	<< "lpopt version 2.1" << endl << endl
			<< "Copyright 2015-2016, Manuel Bichler" << endl
			<< "WWW: <http://dbai.tuwien.ac.at/proj/lpopt/>" << endl
			<< endl
			<< "lpopt is free software: you can redistribute it and/or modify" << endl
			<< "it under the terms of the GNU General Public License as published by" << endl
			<< "the Free Software Foundation, either version 3 of the License, or" << endl
			<< "(at your option) any later version." << endl
			<< endl
			<< "lpopt is distributed in the hope that it will be useful," << endl
			<< "but WITHOUT ANY WARRANTY; without even the implied warranty of" << endl
			<< "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the" << endl
			<< "GNU General Public License for more details." << endl
			<< endl
			<< "You should have received a copy of the GNU General Public License" << endl
			<< "along with lpopt.  If not, see <http://www.gnu.org/licenses/>." <<endl
		;
		exit(EXIT_SUCCESS);
	}

	if(fOpt)
	{
		file.open(filename, ios::in);
		if(!file.is_open()) 
		{ 
			cerr << sProgramName << ": Could not find file '" << filename << "'!" << endl;
			exit(EXIT_FAILURE); 
		}
		stream = &file;
	}

	/* End Argument Handling */

	srand(seed);

	// if(sOpt) cerr << "Using seed: " << seed << endl;

	/* Decomposition Algorithm Selection */

	if(dOpt)
	{
		parser::Driver d(stream);
		Program *p = d.parse();
		p->write(cout);
		delete p;
	}
	else
	{
		LogicProgramOptimizationManager manager(stream, decomp, iOpt);
		unique_ptr<Program> p(manager.optimize());
		p->write(cout);

		if (lOpt)
		{
			ofstream logf;
	                logf.open(logfileName);
        	        if(!logf.is_open())
                	{
                        	cerr << sProgramName << ": Could not open file '" << filename << "'!" << endl;
	                        exit(EXIT_FAILURE);
        	        }
			logf << "width: " << manager.getMaxWidth() << endl;
		}
		
	}
	exit(EXIT_SUCCESS);
}

static void usage()
{
	cout 	<< "Usage: " 
		<< sProgramName 
		<< " [-idv] [-s <seed>] [-f <file>] [-h <alg>] [-l <file>]" << endl 
		<< "\t-d\t\tdumb: do not perform optimization" << endl
		<< "\t-i\t\tignore head variables when decomposing" << endl
		<< "\t-h alg\t\tdecomposition algorithm, one of {mcs, mf, miw (default), noa}" << endl
		<< "\t-s seed\t\tinitialize random number generator with seed." << endl
		<< "\t-f file\t\tthe file to read from (default is stdin)" << endl
		<< "\t-l file\t\toutput infos (treewidth) to file" << endl
		<< "\t-v\t\tprint version, copyright and licence information" << endl
		; // TODO: flag for printing #hide directives for temp. predicates

	exit(EXIT_FAILURE);
}

