/*
Copyright 2015-2016, Manuel Bichler
WWW: <http://dbai.tuwien.ac.at/proj/lpopt/>

This file is part of lpopt.

lpopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

lpopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with lpopt.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "HypergraphableStatement.hpp"
#include <list>

using namespace std;
using namespace lpopt;

static inline vector<const Variable *> getVariables(const Statement& stmt)
{
	vector<const Variable *> varlist;
	varset* globalVars = stmt.variables();
	varlist.insert(varlist.end(), globalVars->begin(), globalVars->end());
	delete globalVars;
	return varlist;
}

HypergraphableStatement::HypergraphableStatement(const Statement& toWrap, bool ignoreHeadVariables, StringGenerator& tempPredicateNameGenerator) : wrapped(&toWrap), stringGenerator(&tempPredicateNameGenerator), ignoreHeadVariables(ignoreHeadVariables), varlist(getVariables(toWrap)), reverseNameMap()
{
}

HypergraphableStatement::~HypergraphableStatement()
{
}

IHypergraph* HypergraphableStatement::toHypergraph(const htd::LibraryInstance& htd)
{
	Hypergraph* hg = new Hypergraph(&htd);
	firstVertex = hg->addVertices(varlist.size()); // add vertices
	// build reverse name map:
	for (auto it = varlist.begin(); it != varlist.end(); ++it)
		reverseNameMap[*it] = firstVertex + it - varlist.begin();
	// edges:
	set<varset*>* elementVars = ignoreHeadVariables ? wrapped->bodyVars() : wrapped->elementVars();
	for (auto it1 = elementVars->begin(); it1 != elementVars->end(); ++it1)
	{
		vector<vertex_t> edge;
		for (auto it2 = (*it1)->begin(); it2 != (*it1)->end(); ++it2)
			edge.push_back(getIdOfVariable(**it2));
		hg->addEdge(edge);
		delete *it1;
	}
	delete elementVars;
	return hg;
}

const Variable * HypergraphableStatement::getVariableById(vertex_t vertex) const
{
	return varlist[vertex - firstVertex];
}

vertex_t HypergraphableStatement::getIdOfVariable(const Variable& var) const
{
	return reverseNameMap.at(&var);
}

bool HypergraphableStatement::ignoreHead() const
{
	return ignoreHeadVariables;
}

const Statement& HypergraphableStatement::stmt() const
{
	return *wrapped;
}


void HypergraphableStatement::setHeadAsRoot(IMutableTreeDecomposition& decomp) const
{
	if (ignoreHeadVariables) return;

	unique_ptr<varset> headVars(wrapped->headVars());
	list<vertex_t> headVertices;
	for (varset::iterator it = headVars->begin();  it != headVars->end(); ++it)
		headVertices.push_back(this->getIdOfVariable(*(*it)));
	headVertices.sort();

	decomp.makeRoot(findContainingBag(decomp, decomp.root(), headVertices.begin(), headVertices.end()));
}

template <class VertexIterator>
vertex_t HypergraphableStatement::findContainingBag(ITreeDecomposition& decomp, vertex_t start, VertexIterator begin, VertexIterator end)
{
	list<vertex_t> bag(decomp.bagContent(start).begin(), decomp.bagContent(start).end());
	bag.sort();
	if(std::includes(bag.begin(), bag.end(), begin, end))
		return start;
	ConstCollection<vertex_t> children = decomp.children(start);
	for(ConstIterator<vertex_t> it = children.begin(); it != children.end(); ++it)
	{
		vertex_t recRes = findContainingBag(decomp, *it, begin, end);
		if(recRes != 0)
			return recRes;
	}
	return 0;
}

StringGenerator& HypergraphableStatement::gen()
{
	return *stringGenerator;
}

RootingOperation* HypergraphableStatement::createHeadRootingOperation(const LibraryInstance * managementInstance) const
{
	return new RootingOperation(*this, managementInstance);
}

RootingOperation::RootingOperation(const HypergraphableStatement& stmt, const LibraryInstance * managementInstance) : stmt(&stmt), mgm(managementInstance)
{
}

RootingOperation::~RootingOperation()
{
}

void RootingOperation::apply(const htd::IMultiHypergraph & /*graph*/, htd::IMutableTreeDecomposition & decomposition) const
{
	stmt->setHeadAsRoot(decomposition);	
}

void RootingOperation::apply(const htd::IMultiHypergraph & graph, htd::IMutableTreeDecomposition & decomposition, const std::vector<htd::vertex_t> & /*relevantVertices*/, std::vector<htd::vertex_t> & /*createdVertices*/, std::vector<htd::vertex_t> & /*removedVertices*/) const
{
	apply(graph, decomposition);
}

void RootingOperation::apply(const htd::IMultiHypergraph & graph, htd::IMutableTreeDecomposition & decomposition, const std::vector<htd::ILabelingFunction *> & /*labelingFunctions*/) const
{
	apply(graph, decomposition);
}

void RootingOperation::apply(const htd::IMultiHypergraph & graph, htd::IMutableTreeDecomposition & decomposition, const std::vector<htd::vertex_t> & /*relevantVertices*/, const std::vector<htd::ILabelingFunction *> & /*labelingFunctions*/, std::vector<htd::vertex_t> & /*createdVertices*/, std::vector<htd::vertex_t> & /*removedVertices*/) const
{
	apply(graph, decomposition);
}

bool RootingOperation::isLocalOperation(void) const
{
	return false;
}

bool RootingOperation::createsTreeNodes(void) const
{
	return false;
}

bool RootingOperation::removesTreeNodes(void) const
{
	return false;
}

bool RootingOperation::modifiesBagContents(void) const
{
	return false;
}

bool RootingOperation::createsSubsetMaximalBags(void) const
{
	return false;
}

bool RootingOperation::createsLocationDependendLabels(void) const
{
	return false;
}

RootingOperation * RootingOperation::clone(void) const
{
	return new RootingOperation(*stmt, mgm);
}
RootingOperation * RootingOperation::cloneTreeDecompositionManipulationOperation(void) const
{
	return clone();
}
const LibraryInstance * RootingOperation::managementInstance(void) const HTD_NOEXCEPT
{
	return mgm;
}
void RootingOperation::setManagementInstance(const LibraryInstance * const manager)
{
	mgm = manager;
}
const RootingOperation * RootingOperation::cloneDecompositionManipulationOperation(void) const
{
	return clone();
}



